module Humidifier
  module Reservoir
    # A CLI for running commands to manipulate the stacks that Reservoir knows
    # about.
    class CLI < Thor
      class_option :aws_profile, desc: 'The AWS profile to authenticate with',
                                 aliases: ['-p']

      class_option :debug, desc: 'Sets up debug mode', aliases: ['-d']
      class_around :safe_execute

      desc 'change [?stack]', 'Create changesets for one or all stacks'
      def change(name = nil)
        stack_names = stack_names_from(name)
        authorize

        stack_names.each do |stack_name|
          stack = Stack.new(stack_name)
          puts "Creating a changeset for #{stack.stack_name}"
          stack.create_change_set
        end
      end

      desc 'deploy [?stack] [*parameters]', 'Update one or all stacks'
      option :wait, desc: 'Wait for the stack to create/update',
                    type: :boolean, default: false
      option :prefix, desc: 'The prefix to use for the stack'
      def deploy(name = nil, *parameters)
        stack_names = stack_names_from(name)
        authorize

        stack_names.each do |stack_name|
          stack = Stack.new(stack_name, prefix: options[:prefix])
          puts "Deploying #{stack.stack_name}"
          stack.deploy(options[:wait], parameters_from(parameters))
        end
      end

      desc 'display [stack] [?pattern]',
           'Display the CloudFormation JSON for a given stack'
      def display(name, pattern = nil)
        puts Stack.new(name, pattern: pattern && /#{pattern}/i).to_cf
      end

      desc 'upload [?stack]', 'Upload one or all stacks to S3'
      def upload(name = nil)
        stack_names = stack_names_from(name)
        authorize

        stack_names.each do |stack_name|
          stack = Stack.new(stack_name)
          puts "Uploading #{stack.stack_name}"
          stack.upload
        end
      end

      desc 'validate [?stack]',
           'Validate that one or all stacks are valid with CloudFormation'
      def validate(name = nil)
        stack_names = stack_names_from(name)
        authorize

        print 'Validating... '

        if stack_names.all? { |stack_name| Stack.new(stack_name).valid? }
          puts 'Valid.'
        else
          puts 'Invalid.'
        end
      end

      no_commands do
        def authorize
          return unless options[:aws_profile]
          Aws.config[:credentials] =
            Aws::SharedCredentials.new(profile_name: options[:aws_profile])
        end

        def parameters_from(opts)
          opts.map do |opt|
            key, value = opt.split('=')
            { parameter_key: key, parameter_value: value }
          end
        end

        def safe_execute
          yield
        rescue Error => error
          raise error if options[:debug]
          puts error.message
          exit 1
        end

        def stack_names_from(name)
          name ? [name] : Reservoir.stacks
        end
      end
    end
  end
end
